/*
** $Id: main.c,v 1.6 2013/02/25 02:57:48 lpitcher Exp $
**
** Purpose:	This program manages a semaphore file to track the number
**		of outstanding scripts executing. When the program detects
**		the start of the first execution of the target script, it
**		will, prior to executing the target script, run one or more
**		resource-creation scripts. When the program detects the end
**		of the final execution of the target script, it will, after
**		the execution of the target script, run one or more
**		resource-destruction scripts.
**
**		This behaviour permits automatic external creation of
**		required resources before the first target script executes,
**		and automatic cleanup of those resources after the last
**		target script completes.
**
**		The intended use of this program is as a hashbang shell
**		interpreter embedded directly in various scripts. The
**		resource creation and cleanup scripts are invoked to (for
**		example) mount and umount external filesystems, pre and
**		post process various data files, etc.
**
** Usage:	<command> resourcelist [/path/to/interpreter] /path/to/target
**
**		where resourcelist consists of
**			resource[,resourcelist]
*/

#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <libgen.h>
#include <string.h>

#include "config.h"
#include "elist.h"
#include "resources.h"
#include "process.h"
#include "exec_shell.h"

char *PgmName,		/* name of this binary, for error messages */
     *JobBase,		/* name of JOB's base directory, for message logging */
     *JobScript;	/* name of JOB's target script, for message logging */

static int ParseArgs(char *[3], int, char **);
static int doList(elist *);

/*
** main()	first level process:
**		- validate arguments and build list of all resources to be managed
**		- fork off second level (child) process to manage the resource list
**		- terminate with returncode based on success of argops.handling, child launch
*/
int main(int argc, char **argv)
{
  int rc = 0;

  char *options[3];
  #define RESOURCELIST	0
  #define INTERPRETER	1
  #define TARGETSCRIPT	2

  PgmName = basename(strdup(argv[0]));

  if (ParseArgs(options,argc,argv) == 0)
  {
    elist *targetsList;

    /* derive pgmname, launch directory name, script name for stderr & syslogging */
    JobBase   = basename(dirname(strdup(options[TARGETSCRIPT])));
    JobScript = basename(strdup(options[TARGETSCRIPT]));

    /* Parent logs with name of source directory */
    openlog(JobBase,LOG_CONS,SYSLOG_FACILITY);

    /* validate the resource names, build the list */
    if (targetsList = buildList(options[RESOURCELIST],options[INTERPRETER],options[TARGETSCRIPT]))
    {
         pid_t worker;

         switch (worker = fork())
         {
           case 0: /* child */
             {
		openlog(JobScript,LOG_CONS|LOG_PID,SYSLOG_FACILITY);
		syslog(SYSLOG_LEVEL_MESSAGE,"JOB Begins");

		rc = doList(targetsList);	/* returncode from target, or RUNERROR */
                if (rc == RUNERROR) rc = FAILED_EXEC;

		syslog(SYSLOG_LEVEL_MESSAGE,"JOB Status %d",rc);
             }
             break;

           case -1: /* parent, fork failed */
             syslog(SYSLOG_LEVEL_ERROR,"JOB %s Failed to launch",JobScript);
             fprintf(stderr,"JOB %s Failed to launch\n",JobScript);
             rc = FAILED_LAUNCH;
             break;

           default:
             syslog(SYSLOG_LEVEL_MESSAGE,"JOB %s[%lu] Launched",JobScript,(unsigned long)worker);
             break;
         }
    }
    else rc = FAILED_LAUNCH;
    closelog();
  }
  else rc = FAILED_ARGS;

  return rc;
}

/* ParseArgs()	extracts resourcelist, interpreter, and script
**		from the argv argument list given to main
** Updates:	an character pointer array, to point to extracted arguments
** Returns:	0 indicating success, or
**		1 indicating failure
*/
static int ParseArgs(char *options[3], int argc, char **argv)
{
  int result = 1;
  char *p;

  switch (argc)
  {
    case 0:
    case 1:
    case 2:
      fprintf(stderr,"%s: missing commandline arguments\n",PgmName);
      break;
    
    case 3:
      /*
      ** command "resourcelist interpreter" script
      **   OR
      ** command resourcelist script
      */

      for (p = argv[1]; (*p == ' ') || (*p == '\t'); ++p)
      { /* skip leading whitespace */ }

      if (*p) /* Not at argv[1] end-of-string yet */
      {
        options[RESOURCELIST] = p;

        /* find whitespace separating resourcelist and target */
        while (*p && (*p != ' ') && (*p != '\t')) ++p;

        if (*p) /* Not at argv[1] end-of-string yet */
        {
          *p = 0; /* terminate resourcelist string */

          for (++p; (*p == ' ') || (*p == '\t') ; ++p)
          { /* skip intermediate whitespace */ }

          /*
          ** At this point, p points to either
          **   the first character of the INTERPRETER path, or
	  **   a \0 flagging the end of argv[1]
          ** We interpret the \0 as an /unspecified/ INTERPRETER
          **  and let later logic substitute the default path
          */
        }
        else break; /* all-blank or empty argv[1] is an error */

        /* if p points to end-of-string, use the default */
        options[INTERPRETER] = (*p) ? p : strdup("/bin/sh");

        options[TARGETSCRIPT] = argv[2];	/* SCRIPT */
        result = 0;
      }
      break;

    case 4:
      /*
      ** command resourcelist interpreter script
      */
      options[RESOURCELIST] = argv[1];
      options[INTERPRETER]  = argv[2];
      options[TARGETSCRIPT] = argv[3];
      result = 0;
      break;

    default:
      fprintf(stderr,"%s: too many commandline arguments\n",PgmName);
      break;
  }

  if (result != 0)
    fprintf(stderr,
            "Usage: %s resourcename[,resourcename...] [interpreter] script\n",
            PgmName);

  return result;
}

/*
** doList()	launch required preprocessors, target script, and postprocessors
**
** Accepts:	a pointer to the top struct eList node in a list of nodes
**		describing the required resources and target script
**
** Updates:	semaphore files for each active resource
**		syslog with stdout, stderr from each resource processor
**		syslog with stdout, stderr from target script
**		closes semaphore files, updates related eList node
**
** Returns:	returncode from the target script, or RUNERROR
*/
static int doList(elist *this)
{
  int rc;

  if (this->ExecNext != NULL)	/* preprocessor/postprocessor entry */
  {
    if (doPrep(this->ExecTarget,this->ExecParm.ExecFile) == PROC_WORKED)
    {
      rc = doList(this->ExecNext);
      doPost(this->ExecTarget,this->ExecParm.ExecFile);
    }
    else rc = RUNERROR;
    close(this->ExecParm.ExecFile); this->ExecParm.ExecFile = -1;
  }
  else rc = doExec(this->ExecParm.ExecInterp,this->ExecTarget);

  return rc;
}


/*
** Copyright Lew Pitcher (lew.pitcher@digitalfreehold.ca) February, 2013
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program (see the licence.txt file); if not, write to the Free
** Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
** 02110-1301 USA.
*/
